unit gm;

{$MODE OBJFPC}{$H+}

interface

uses
  Crt, snk;


type
  { game class }
  TGame = class(TObject)
  private
    FSnake: TSnake;
    FGold: TSnakeModuleCoords;
  private
    procedure DrawLevelBounds();
    procedure DrawSnake();
    procedure DrawGold();
    procedure ClearSnakeTail();
  private
    function CanMove(ADirection: TSnakeMoveDirection): Boolean;
    function IsModuleInLevel(APosition: TSnakeModuleCoords): Boolean;
    function IsGoldModule(APosition: TSnakeModuleCoords): Boolean;
    function IsFreeLevelModule(APosition: TSnakeModuleCoords): Boolean;
    function MoveDirectionToModuleCoords(ADirection: TSnakeMoveDirection): TSnakeModuleCoords;
  private
    procedure SetNewGoldPosition();
  public
    constructor Create();
    destructor Destroy(); override;
  public
    procedure Play();
  end;


implementation


const
  { level bounds }
  LEVEL_LEFT_BOUND   = TCrtCoord(7);
  LEVEL_TOP_BOUND    = TCrtCoord(7);
  LEVEL_RIGHT_BOUND  = TCrtCoord(73);
  LEVEL_BOTTOM_BOUND = TCrtCoord(21);

const
  { snake default length }
  SNAKE_DEFAULT_LENGTH = TSnakeLength(3);
  { snake default coordinates }
  SNAKE_DEFAULT_POSITION_X: TSnakeModuleCoords = (
    X: LEVEL_LEFT_BOUND + SNAKE_DEFAULT_LENGTH + 2;
    Y: LEVEL_BOTTOM_BOUND - 2;
  );

const
  { gold symbol }
  GOLD_SYMBOL = Char(#254);


constructor TGame.Create();
begin
  inherited Create();
  Randomize();

  { hide cursor }
  CursorOff();
  { create snake }
  FSnake := TSnake.Create(SNAKE_DEFAULT_LENGTH, SNAKE_DEFAULT_POSITION_X);
end;

destructor TGame.Destroy();
begin
  { destroy snake }
  FSnake.Free();
  { show cursor }
  CursorOn();

  inherited Destroy();
end;

procedure TGame.DrawLevelBounds();

  procedure DrawHorzBound(AX, AY, AWidth: TCrtCoord);
  begin
    { move cursor and draw horizontal bound }
    GotoXY(AX, AY);
    Write(StringOfChar(Char(smtHorzizontal), AWidth));
  end;

  procedure DrawVertBound(AX, AY, AHeight: TCrtCoord);
  var
    I: TCrtCoord;
  begin
    { move cursor and draw vertical bound }
    for I := 1 to AHeight - 1 do
    begin
      GotoXY(AX, AY + I);
      Write(Char(smtVertical));
    end;
  end;

  procedure DrawBoundCorner(AX, AY: TCrtCoord; ACorner: Char);
  begin
    { move cursor and draw bound corner }
    GotoXY(AX, AY);
    Write(ACorner);
  end;

begin
  { set the bounds color }
  TextColor(White);

  { draw horizontal bounds }
  DrawHorzBound(LEVEL_LEFT_BOUND + 1, LEVEL_TOP_BOUND,    LEVEL_RIGHT_BOUND - LEVEL_LEFT_BOUND - 1);
  DrawHorzBound(LEVEL_LEFT_BOUND + 1, LEVEL_BOTTOM_BOUND, LEVEL_RIGHT_BOUND - LEVEL_LEFT_BOUND - 1);
  { draw vertical bounds }
  DrawVertBound(LEVEL_LEFT_BOUND,  LEVEL_TOP_BOUND, LEVEL_BOTTOM_BOUND - LEVEL_TOP_BOUND);
  DrawVertBound(LEVEL_RIGHT_BOUND, LEVEL_TOP_BOUND, LEVEL_BOTTOM_BOUND - LEVEL_TOP_BOUND);
  { draw bound corners }
  DrawBoundCorner(LEVEL_LEFT_BOUND,  LEVEL_TOP_BOUND,    Char(smtCornerRightDown));
  DrawBoundCorner(LEVEL_RIGHT_BOUND, LEVEL_TOP_BOUND,    Char(smtCornerLeftDown));
  DrawBoundCorner(LEVEL_LEFT_BOUND,  LEVEL_BOTTOM_BOUND, Char(smtCornerRightUp));
  DrawBoundCorner(LEVEL_RIGHT_BOUND, LEVEL_BOTTOM_BOUND, Char(smtCornerLeftUp));
end;

procedure TGame.DrawSnake();
var
  smrToken: TSnakeModuleRec;
begin
  { set the snake head and tail color }
  TextColor(Yellow);

  { draw snake head module }
  smrToken := FSnake[FSnake.Length - 1];
  { move cursor and draw snake head }
  GotoXY(smrToken.Coords.X, smrToken.Coords.Y);
  Write(Char(smrToken.ModuleType));

  { draw snake tail module }
  smrToken := FSnake[0];
  { move cursor and draw snake tail }
  GotoXY(smrToken.Coords.X, smrToken.Coords.Y);
  Write(Char(smrToken.ModuleType));

  { set the snake body color }
  TextColor(Green);

  { draw snake neck module }
  smrToken := FSnake[FSnake.Length - 2];
  { move cursor and draw snake neck }
  GotoXY(smrToken.Coords.X, smrToken.Coords.Y);
  Write(Char(smrToken.ModuleType));
end;

procedure TGame.DrawGold();
begin
  { set the gold color }
  TextColor(Yellow);

  { move cursor and draw gold }
  GotoXY(FGold.X, FGold.Y);
  Write(GOLD_SYMBOL);
end;

procedure TGame.ClearSnakeTail();
var
  smrToken: TSnakeModuleRec;
begin
  { get the snake tail module record }
  smrToken := FSnake[0];
  { move cursor and clear snake tail}
  GotoXY(smrToken.Coords.X, smrToken.Coords.Y);
  Write(#32);
end;

function TGame.CanMove(ADirection: TSnakeMoveDirection): Boolean;
var
  smcHead: TSnakeModuleCoords;
begin
  { get head of snake coordinates }
  smcHead := FSnake.HeadCoords;

  { modify the temp head coordinates }
  case ADirection of
    smdUp:    Dec(smcHead.Y);
    smdDown:  Inc(smcHead.Y);
    smdLeft:  Dec(smcHead.X);
    smdRight: Inc(smcHead.X);
  end;

  { level area }
  Result := IsModuleInLevel(smcHead);
end;

function TGame.IsModuleInLevel(APosition: TSnakeModuleCoords): Boolean;
begin
  { check the bound colision }
  Result := (APosition.X in [LEVEL_LEFT_BOUND + 1 .. LEVEL_RIGHT_BOUND - 1]) and
            (APosition.Y in [LEVEL_TOP_BOUND + 1  .. LEVEL_BOTTOM_BOUND - 1]);
end;

function TGame.IsGoldModule(APosition: TSnakeModuleCoords): Boolean;
begin
  { check the gold colision }
  Result := (APosition.X = FGold.X) and (APosition.Y = FGold.Y);
end;

function TGame.IsFreeLevelModule(APosition: TSnakeModuleCoords): Boolean;
var
  slToken:  TSnakeLength = 0;
  smrToken: TSnakeModuleRec;
begin
  while slToken < FSnake.Length do
  begin
    { get the snake module record }
    smrToken := FSnake[slToken];

    { same module }
    if (smrToken.Coords.X = APosition.X) and (smrToken.Coords.Y = APosition.Y) then
      Exit(False);

    { increment snake modules token }
    Inc(slToken);
  end;

  Result := True;
end;

function TGame.MoveDirectionToModuleCoords(ADirection: TSnakeMoveDirection): TSnakeModuleCoords;
begin
  { get snake head module coordinates }
  Result := FSnake.HeadCoords;

  { modyfi head module coordinates }
  case ADirection of
    smdUp:    Dec(Result.Y);
    smdDown:  Inc(Result.Y);
    smdLeft:  Dec(Result.X);
    smdRight: Inc(Result.X);
  end;
end;

procedure TGame.SetNewGoldPosition();
var
  smcGold: TSnakeModuleCoords;
  boolValid: Boolean;
begin
  repeat
    { set the valid flag }
    boolValid := False;

    { get new gold coordinates }
    smcGold.X := Random(LEVEL_RIGHT_BOUND  - LEVEL_LEFT_BOUND - 1) + LEVEL_LEFT_BOUND + 1;
    smcGold.Y := Random(LEVEL_BOTTOM_BOUND - LEVEL_TOP_BOUND  - 1) + LEVEL_TOP_BOUND  + 1;

    { validate coordinates }
    boolValid := IsFreeLevelModule(smcGold);
  until boolValid;

  { set the new gold coordinates }
  FGold := smcGold;
end;

procedure TGame.Play();
const
  { move keys }
  KEY_UP    = Char('H');
  KEY_DOWN  = Char('P');
  KEY_LEFT  = Char('K');
  KEY_RIGHT = Char('M');
  { options keys }
  KEY_RESET = Char('R');
  KEY_EXIT  = Char('E');

  function InputKeyToDirection(const AKey: Char): TSnakeMoveDirection;
  begin
    { translate input key }
    case AKey of
      KEY_UP:    Result := smdUp;
      KEY_DOWN:  Result := smdDown;
      KEY_LEFT:  Result := smdLeft;
      KEY_RIGHT: Result := smdRight;
    end;
  end;

const
  INPUT_MOVE_KEYS: set of Char = [KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT];
  INPUT_OPT_KEYS:  set of Char = [KEY_RESET, KEY_EXIT];
var
  chrInput: Char;
  smdInput: TSnakeMoveDirection;
  boolEat: Boolean;
begin
  ClrScr;
  { draw level bounds }
  DrawLevelBounds();
  { set the first gold coordinates }
  SetNewGoldPosition();

  { main game loop }
  while True do
  begin
    { draw snake at current position }
    DrawSnake();
    { draw gold at current position }
    DrawGold();

    repeat
      { get the input key }
      chrInput := UpCase(ReadKey());
    until (chrInput in INPUT_MOVE_KEYS) or (chrInput in INPUT_OPT_KEYS);

    { is move key }
    if chrInput in INPUT_MOVE_KEYS then
    begin
      { translate key }
      smdInput := InputKeyToDirection(chrInput);

      { validate move }
      if FSnake.CanMove(smdInput) and Self.CanMove(smdInput) then
      begin
        { gold module }
        boolEat := IsGoldModule(MoveDirectionToModuleCoords(smdInput));

        { clear snake tail }
        ClearSnakeTail();
        { move snake }
        FSnake.Move(smdInput, boolEat);

        { new gold coordinates }
        if boolEat then
          SetNewGoldPosition();
      end;
    end
    else
      { handle an option }
      case chrInput of
        { reset snake position }
        KEY_RESET:
            begin
              ClrScr;
              { draw level bounds }
              DrawLevelBounds();
              { reset snake position }
              FSnake.Reset(SNAKE_DEFAULT_LENGTH, SNAKE_DEFAULT_POSITION_X);
              { reset gold position }
              SetNewGoldPosition();
            end;
        { exit game loop }
        KEY_EXIT: Exit;
      end;
  end;

  ClrScr;
end;


end.

