unit snk;

{$MODE OBJFPC}{$H+}{$PACKENUM 1}

interface

uses
  Crt;


type
  TSnakeLength = UInt16;

type
  { snake single module type enumeration }
  TSnakeModuleType = (
    smtVertical        = 186,
    smtCornerLeftDown  = 187,
    smtCornerLeftUp    = 188,
    smtCornerRightUp   = 200,
    smtCornerRightDown = 201,
    smtHorzizontal     = 205,
    smtHead            = 219
  );

type
  { snake single module coordinates }
  TSnakeModuleCoords = record
    X, Y: TCrtCoord;
  end;

type
  { snake single module record }
  TSnakeModuleRec = record
    ModuleType: TSnakeModuleType;
    Coords: TSnakeModuleCoords;
  end;

type
  { array of snake modules }
  TSnakeModulesArr = array of TSnakeModuleRec;

type
  { snake move direction enumeration }
  TSnakeMoveDirection = (smdUp, smdDown, smdLeft, smdRight);

type
  { snake class }
  TSnake = class(TObject)
  private
    FModules: TSnakeModulesArr;
    FLength: TSnakeLength;
    FCurrDirection: TSnakeMoveDirection;
  private
    procedure BuildSnake(ALength: TSnakeLength; APosition: TSnakeModuleCoords);
  private
    function GetModule(AIndex: TSnakeLength): TSnakeModuleRec;
    function GetHeadCoords(): TSnakeModuleCoords;
    function IsOppositeDirection(ADirection: TSnakeMoveDirection): Boolean;
    function IsFreeModulePosition(APosition: TSnakeModuleCoords): Boolean;
    function CoordsRelationshipToMoveDirection(AEarlierIndex, ALaterIndex: TSnakeLength): TSnakeMoveDirection;
  public
    constructor Create(ALength: TSnakeLength; APosition: TSnakeModuleCoords);
    destructor Destroy(); override;
  public
    function CanMove(ADirection: TSnakeMoveDirection): Boolean;
    procedure Move(ADirection: TSnakeMoveDirection; AEat: Boolean);
    procedure Reset(ALength: TSnakeLength; APosition: TSnakeModuleCoords);
  public
    property Length: TSnakeLength read FLength;
    property HeadCoords: TSnakeModuleCoords read GetHeadCoords;
  public
    property Module[AIndex: TSnakeLength]: TSnakeModuleRec read GetModule; default;
  end;


implementation


constructor TSnake.Create(ALength: TSnakeLength; APosition: TSnakeModuleCoords);
begin
  inherited Create();

  { build snake }
  BuildSnake(ALength, APosition);
  { set the current snake move direction }
  FCurrDirection := smdRight;
end;

destructor TSnake.Destroy();
begin
  inherited Destroy();
end;

procedure TSnake.BuildSnake(ALength: TSnakeLength; APosition: TSnakeModuleCoords);
var
  slToken:  TSnakeLength       absolute ALength;
  smcToken: TSnakeModuleCoords absolute APosition;
begin
  FLength := ALength;

  { set the length of snake }
  SetLength(FModules, slToken);
  Dec(slToken);

  { fill snake head module }
  FModules[slToken].ModuleType := smtHead;
  FModules[slToken].Coords     := smcToken;

  { fill snake tail modules }
  repeat
    Dec(slToken);
    Dec(smcToken.X);
    FModules[slToken].ModuleType := smtHorzizontal;
    FModules[slToken].Coords     := smcToken;
  until slToken = 0;
end;

function TSnake.GetModule(AIndex: TSnakeLength): TSnakeModuleRec;
begin
  { return module from modules array }
  Result := FModules[AIndex];
end;

function TSnake.GetHeadCoords(): TSnakeModuleCoords;
begin
  { return the head of snake coordinates }
  Result := FModules[FLength - 1].Coords;
end;

function TSnake.IsOppositeDirection(ADirection: TSnakeMoveDirection): Boolean;
var
  smdMoves: set of TSnakeMoveDirection;
begin
  { check the opposite state }
  smdMoves := [FCurrDirection, ADirection];
  Result   := (smdMoves = [smdUp, smdDown]) or (smdMoves = [smdLeft, smdRight]);
end;

function TSnake.IsFreeModulePosition(APosition: TSnakeModuleCoords): Boolean;
var
  intToken: Integer;
begin
  { check the modules colision }
  for intToken := FLength - 4 downto 0 do
    if (FModules[intToken].Coords.X = APosition.X) and (FModules[intToken].Coords.Y = APosition.Y) then
    begin
      Result := False;
      Exit;
    end;

  Result := True;
end;

function TSnake.CoordsRelationshipToMoveDirection(AEarlierIndex, ALaterIndex: TSnakeLength): TSnakeMoveDirection;
const
  HORZ_MOVE_DIRECTION: array [Boolean] of TSnakeMoveDirection = (smdLeft, smdRight);
  VERT_MOVE_DIRECTION: array [Boolean] of TSnakeMoveDirection = (smdUp, smdDown);
var
  smcEarlier, smcLater: TSnakeModuleCoords;
begin
  smcEarlier  := FModules[AEarlierIndex].Coords;
  smcLater := FModules[ALaterIndex].Coords;

  if smcEarlier.X = smcLater.X then
    Result := VERT_MOVE_DIRECTION[smcEarlier.Y > smcLater.Y]
  else
    Result := HORZ_MOVE_DIRECTION[smcEarlier.X > smcLater.X];
end;

function TSnake.CanMove(ADirection: TSnakeMoveDirection): Boolean;
var
  smcHead: TSnakeModuleCoords;
begin
  { get head of snake coordinates }
  smcHead := GetHeadCoords();

  { opposite direction }
  if IsOppositeDirection(ADirection) then Exit(False);

  { modify the temp head coordinates }
  case ADirection of
    smdUp:    Dec(smcHead.Y);
    smdDown:  Inc(smcHead.Y);
    smdLeft:  Dec(smcHead.X);
    smdRight: Inc(smcHead.X);
  end;

  { free module }
  Result := IsFreeModulePosition(smcHead);
end;

procedure TSnake.Move(ADirection: TSnakeMoveDirection; AEat: Boolean);
var
  slToken: TSnakeLength = 0;
  smdHeadToNeck, smdNeckToNext: TSnakeMoveDirection;
begin
  { update the current move direction }
  FCurrDirection := ADirection;

  { eat or not }
  if AEat then
  begin
    { increase snake length }
    Inc(FLength);
    SetLength(FModules, FLength);
    { set the modules token value }
    slToken := FLength - 1;
    { set the new head data }
    FModules[slToken].ModuleType := smtHead;
    FModules[slToken].Coords := FModules[slToken - 1].Coords;
  end
  else
    { rewrite body modules coordinates }
    while slToken < FLength - 1 do
    begin
      FModules[slToken].Coords := FModules[slToken + 1].Coords;
      Inc(slToken);
    end;

  { update the new head module coordinates }
  case ADirection of
    smdUp:    Dec(FModules[slToken].Coords.Y);
    smdDown:  Inc(FModules[slToken].Coords.Y);
    smdLeft:  Dec(FModules[slToken].Coords.X);
    smdRight: Inc(FModules[slToken].Coords.X);
  end;

  if AEat then
    slToken := FLength - 2
  else
  begin
    slToken := 0;

    { rewrite body modules types }
    while slToken < FLength - 2 do
    begin
      FModules[slToken].ModuleType := FModules[slToken + 1].ModuleType;
      Inc(slToken);
    end;
  end;

  { get the head, neck and next modules coords relationship }
  smdHeadToNeck := CoordsRelationshipToMoveDirection(slToken + 1, slToken);
  smdNeckToNext := CoordsRelationshipToMoveDirection(slToken, slToken - 1);

  { update the neck module type }

  { straight vertical }
  if ((smdHeadToNeck = smdUp)   and (smdNeckToNext = smdUp)) or
     ((smdHeadToNeck = smdDown) and (smdNeckToNext = smdDown)) then
  begin
    FModules[slToken].ModuleType := smtVertical;
    Exit;
  end;

  { straight horizontal }
  if ((smdHeadToNeck = smdLeft)  and (smdNeckToNext = smdLeft)) or
     ((smdHeadToNeck = smdRight) and (smdNeckToNext = smdRight)) then
  begin
    FModules[slToken].ModuleType := smtHorzizontal;
    Exit;
  end;

  { up and left corner & down and right corner}
  if ((smdHeadToNeck = smdLeft) and (smdNeckToNext = smdUp)) or
     ((smdHeadToNeck = smdDown) and (smdNeckToNext = smdRight)) then
  begin
    FModules[slToken].ModuleType := smtCornerLeftDown;
    Exit;
  end;

  { up and right corner & down and left corner }
  if ((smdHeadToNeck = smdRight) and (smdNeckToNext = smdUp)) or
     ((smdHeadToNeck = smdDown)  and (smdNeckToNext = smdLeft)) then
  begin
    FModules[slToken].ModuleType := smtCornerRightDown;
    Exit;
  end;

  { down and left corner & up and right corner }
  if ((smdHeadToNeck = smdLeft) and (smdNeckToNext = smdDown)) or
     ((smdHeadToNeck = smdUp)   and (smdNeckToNext = smdRight)) then
  begin
    FModules[slToken].ModuleType := smtCornerLeftUp;
    Exit;
  end;

  { down and right corner & up and left corner }
  if ((smdHeadToNeck = smdRight) and (smdNeckToNext = smdDown)) or
     ((smdHeadToNeck = smdUp)    and (smdNeckToNext = smdLeft)) then
  begin
    FModules[slToken].ModuleType := smtCornerRightUp;
    Exit;
  end;
end;

procedure TSnake.Reset(ALength: TSnakeLength; APosition: TSnakeModuleCoords);
begin
  { rebuild snake modules }
  BuildSnake(ALength, APosition);
  { update the current snake move direction }
  FCurrDirection := smdRight;
end;


end.
